/*

ARMiner - Association Rules Miner
Copyright (C) 2000  UMass/Boston - Computer Science Department


This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA


The ARMiner Server was written by Dana Cristofor and Laurentiu
Cristofor.

The ARMiner Client was written by Abdelmajid Karatihy, Xiaoyong Kuang,
and Lung-Tsung Li.


The ARMiner package is currently maintained by Laurentiu Cristofor
(laur@cs.umb.edu).

*/

package wpi.associations.arminer;

import java.io.*;
import java.util.*;
import weka.core.*;

/**

  DBReader.java<P>

  A DBReader is used to read data from a database.<P>
   
*/
/*
  
   This file is a part of the ARMiner project.
   
   (P)1999-2000 by ARMiner Server Team:

   Dana Cristofor
   Laurentiu Cristofor

*/
/*
   This file has been modified to load a databases that store set-valued
   attributes from WEKA Instances
   
   Zack Stoecker-Sylvia 2/14/2002
*/
public class DBReader
{
	// A hashtable of item ID Integers, indexed on Attribute=Value strings
	private Hashtable value_hash;
	// A hashtable of Attribute=Value strings, indexed on item ID Integers
	// Used to pass back to WEKA so it can know what the numbers mean
	private Hashtable number_hash;
	// The instances passed by WEKA
	private Instances instances;
	// Current instance index
	private int current_instance;
	// An array of Itemsets to pass back to WEKA if one or more attributes
	// are required
	private ARMinerItemset[] attribute_itemset;
	// These are the characters to use in a set
	// { and } are included for the beginning and the end, but could be
	// used as delimeters if someone really wanted to.
	// Non-set string attributes should simply be parsed as a single
	// item as long as they do not contain any delimeters.
	private static String SET_DELIMETERS = "&^{}";
	
	/**
	 * Loads a WEKA Instances object and sets the valid items
	 * 
	 * @param		instances	the instances to read itemsets from
	 * @exception	exception	if the instances contain a numeric attribute
	 */
	public DBReader(Instances instances) throws UnsupportedAttributeTypeException
	{
		this.instances = instances;
		value_hash = new Hashtable();
		number_hash = new Hashtable();
		attribute_itemset = new ARMinerItemset[instances.numAttributes()];
		
		Enumeration attribute_enumeration = instances.enumerateAttributes();
		Enumeration value_enumeration;
		Attribute a;
		StringTokenizer st;
		String s;
		
		while (attribute_enumeration.hasMoreElements())
		{
			a = (Attribute)attribute_enumeration.nextElement();

			if (a.type() == Attribute.NUMERIC)
			{
				throw new UnsupportedAttributeTypeException("Numeric attributes are not allowed.");
			}
			else
			{
				value_enumeration = a.enumerateValues();
				attribute_itemset[a.index()] = new ARMinerItemset();
				
				while (value_enumeration.hasMoreElements())
				{
					s = (String)value_enumeration.nextElement();
					
					if (a.type() == Attribute.NOMINAL)
					{
						// Add to number_hash, attribute_itemset, and value_hash
						number_hash.put(new Integer(value_hash.size() + 1),a.name() + "=" + s);
						attribute_itemset[a.index()].addItem(value_hash.size() + 1);
						value_hash.put(a.name() + "=" + s,new Integer(value_hash.size() + 1));
					}
					else // Attribute is a string
					{
						st = new StringTokenizer(s,SET_DELIMETERS);
						
						while (st.hasMoreTokens())
						{
							s = st.nextToken();
							
							// if set value has not already been stored
							if (!value_hash.containsKey(a.name() + "=" + s))
							{
								// Add to number_hash, attribute_itemset, and value_hash
								number_hash.put(new Integer(value_hash.size() + 1),a.name() + "=" + s);
								attribute_itemset[a.index()].addItem(value_hash.size() + 1);
								value_hash.put(a.name() + "=" + s,new Integer(value_hash.size() + 1));
							}
						}
					}
				}
			}
		}
	}
	
	/**
	 * Function to return the number of rows (instances)
	 * 
	 * @return	the number of instances
	 */
	public long getNumRows()
	{
		return instances.numInstances();
	}
	
	/**
	 * Function to return the number of columns (attribute value items)
	 * 
	 * @return	the number of attribute value items
	 */
	public long getNumColumns()
	{
		return value_hash.size();
	}
	
	/**
	 * Function to return the first instance as an ARMinerItemset
	 * 
	 * @return	an ARMinerItemset of the first instance
	 */
	public ARMinerItemset getFirstRow() throws IOException
	{
		current_instance = 0;
		
		return getNextRow();
	}
	
	/**
	 * Function to return the instance at current_instance as an ARMinerItemset
	 * 
	 * @return		an ARMinerItemset of the current instance
	 */
	public ARMinerItemset getNextRow()
	{
		Instance inst = instances.instance(current_instance);
		ARMinerItemset is = new ARMinerItemset();

		Enumeration attribute_enumeration = inst.enumerateAttributes();
		Attribute a;
		StringTokenizer st;
		String s;
		Object o;
		
		while (attribute_enumeration.hasMoreElements())
		{
			a = (Attribute)attribute_enumeration.nextElement();
			
			if (!inst.isMissing(a))
			{
				if (a.type() == Attribute.NOMINAL)
				{
					o = value_hash.get(a.name() + "=" + inst.stringValue(a));
					is.addItem(((Integer)o).intValue());
				}
				else
				{
					st = new StringTokenizer(inst.stringValue(a),SET_DELIMETERS);

					while (st.hasMoreTokens())
					{
						s = st.nextToken();
						o = value_hash.get(a.name() + "=" + s);
						is.addItem(((Integer)o).intValue());
					}
				}
			}
		}
		
		current_instance++;
		return is;
	}
	
	/**
	 * Function to check if there are more rows (instances)
	 * 
	 * @return	true if current_instance < number of instances, false otherwise
	 */
	public boolean hasMoreRows()
	{
		if (current_instance < instances.numInstances())
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	
	/**
	 * Function to return the item ID Integer to Attribute=Value string hashtable
	 * for use in translating back from from numbers to strings by WEKA.
	 * 
	 * @return	a Hashtable where Attribute=Value Strings are keyed by Integers
	 */
	public Hashtable getNumberHash()
	{
		return number_hash;
	}
	
	/**
	 * Function to return an ARMinerItemset containing all items in a particular
	 * attribute.  Used by WEKA to specify constraints to AprioriRules.java.
	 * 
	 * @param		i		the index of the attribute
	 * @return		an ARMinerItemset of items in the given attribute
	 */
	public ARMinerItemset getAttributeItemset(int i)
	{
		return attribute_itemset[i];
	}

  public static void main(String arg[]) {
  }
}
