/*
Copyright (C) 2003  WPI - Computer Science Department - Keith A. Pray

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
USA
*/

/*
   This file is a part of the WPI WEKA Project
   It is a greatly modified version of SET.java from the ARMiner Project.
*/

/*
  Maintainance log started on February 22, 2003

  February 22, 2003  - first modified for indexing itemsets -kap

  March 9, 2003      - removed String event from HashNode since it was only
                       used for printing the tree -kap
*/

package wpi.associations.arminerSequence;

import java.util.*;
import java.io.EOFException;

import wpi.timesequence.*;
import weka.core.Utils;

/**
 * A prefix tree used for indexing itemsets. This implementation
 * assumes that all event items of the same name will be numbered
 * consecutively. If the code that generates the first level of
 * candidate itemsets is changed this condition must be ensured.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version 1.0
 */

public class ItemsetPrefixTree
{

//********************************************************//
// ----- Inner Classes -----
//********************************************************//  

  private class HashNode
  {
    //********************************************************//

    /** Our child nodes */
    Hashtable children;

    /** The itemset this nodes indexes, if any */
    ARMinerItemset itemset;

    //********************************************************//    

    public HashNode()
    {
      children = new Hashtable();

      // don't have an itemset yet
      itemset = null;

    } // END public HashNode()

    //********************************************************//    

    public String toString()
    {
      // changed to use string buffer -kap
      StringBuffer sb = new StringBuffer ( "<children: " );

      sb.append ( children.toString() );
      sb.append ( "itemset: " );
      sb.append ( itemset );
      sb.append ( ">" );

      return ( sb.toString() );
    }

    //********************************************************//    

  } // END private class HashNode

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The root node of the tree */
  private HashNode root;

  /** Tracks the current level for pretty printing of tree */
  private int level = 0;

  /** The number of itemsets indexed */
  private int size = 0;

  /** Specifies debug info level 
   * 0: no debug info
   * 1: input to methods
   * 2: and out from methods
   * 3: and all sorts of stuff 
   */
  private static final int debug = 0;

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Returns the number of itemsets indexed.
   *
   * @return the number of itemsets indexed
   */

  public int size()
  {
    return ( size );

  } // END public int size()

//********************************************************//
  
  /**
   * Create a new empty SET.
   */

  public ItemsetPrefixTree()
  {
    root = new HashNode();

  } // END public ItemsetPrefixTree()

//********************************************************//
  
  /**
   * Add a new itemset in the itemset prefix tree.
   * If there is an existing itemset in the prefix tree that
   * represents the same regular and event items as the
   * itemset specified it will be replaced by the specified
   * itemset.
   *
   * @param itemset  
   * the itemset to be added
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   */

  public void add ( ARMinerItemset itemset )
  {
    if ( itemset == null || itemset.size() == 0 )
    {
      throw new IllegalArgumentException 
	( "ItemsetPrefixTree.add: itemset must be non null and non empty" );
    }

    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ItemsetPrefixTree.add: Making new itemset." );
    case 1: //case 0:
      System.out.println ( "ItemsetPrefixTree.add: itemset = " + itemset );
    } // end switch debug -kap

    Object obj;
    HashNode node;
    HashNode walker = root;
    Integer key;

    // if an item is an event there will be another node we have to deal with
    // we should wait until after we are done with the regular items
    HashNode eventNode = null;
    // and the key will be a string
    String eventKey;
    // to build the event key quickly
    StringBuffer sb = new StringBuffer();

    // for each regular item in the itemset walk the branch of the tree
    // associated with it -kap

    // skip event items but record them
    Integer [] eventItem = new Integer [ itemset.numEventItems() ];
    int numEventItem = 0;

    // start with the first item
    itemset.startItemIteration();

    while ( itemset.hasMoreItems() )
    {
      // the key for looking up the next node in the tree is the next
      // item in the itemset -kap
      key = new Integer ( itemset.getNextItem() );
      
      // check if this item is an event
      if ( itemset.eventName ( key ) != null )
      {
	// save the event item
	eventItem [ numEventItem ] = key;
	numEventItem++;

	// then this is an event item and we should skip it for now
	continue;

      } // end if item is an event

      // if there is already a node for this item go to it -kap
      if ( ( obj = walker.children.get ( key ) ) != null )
      {
	walker = (HashNode)obj;
	
      } // end if we found the node for the item
      else
      {
	// otherwise we must make a new node for this item -kap
	node = new HashNode();
	walker.children.put ( key, node );
	walker = node;
	
	// since we know there are no other nodes from here on
	// we could save time by just creating nodes for the rest
	// of the items rather than checking above. But will we?
	// I guess not right now. -kap
	
      } // end else we have to make a new node

    } // end while we have more regular items left

    // we need the name of the event
    String eventName;

    // for each event item in the itemset walk the branch of the tree
    // associated with it -kap

    for ( numEventItem = 0; numEventItem < eventItem.length; numEventItem++ )
    {
      // the key for looking up the next node in the tree is the next
      // item in the itemset -kap
      key = eventItem [ numEventItem ];
      
      eventName = itemset.eventName ( key );

      // better build the key for it then
      sb.setLength ( 0 );
      sb.append ( itemset.relativeBegin ( key ) );
      sb.append ( TSUtility.TIME_SEQUENCE_DELIMITER );
      sb.append ( itemset.relativeEnd ( key ) );
      eventKey = sb.toString();

      // hrm, then the key is going to the event name instead
      // of the item number

      // if there is already a node for this item go to it -kap
      if ( ( obj = walker.children.get ( eventName ) ) != null )
      {
	walker = (HashNode)obj;

	// we have to check if the event key node exists
	if ( ( eventNode = (HashNode)walker.children.get ( eventKey ) ) 
	     != null )
	{
	  // then we walk to the event node
	  walker = eventNode;
	  
	  switch ( debug ) {
	  case 3:
	    System.out.println ( "ItemsetPrefixTree.add: walked to event node "
				 + key + " : " + eventKey + " : " + walker );
	  } // end switch debug
	  
	} // end if event node exists
	else
	{
	  // we had better build a node for the event then
	  eventNode = new HashNode();

	  // add this new event node with the event key
	  walker.children.put ( eventKey, eventNode );
	    
	  // walk to it baby
	  walker = eventNode;
	    
	} // end else we have to make the event node
	
      } // end if we found the node for the item
      else
      {
	// otherwise we must make a new node for this item -kap
	node = new HashNode();

	// remeber to use event name as the key
	walker.children.put ( eventName, node );
	walker = node;
	
	// we also have to make an event node then

	// well, now we have to make another node and add it to
	// the children of the node we just made above
	eventNode = new HashNode();
	
	// add this node with the event key
	node.children.put ( eventKey, eventNode );
	
	// now walk to it baby
	walker = eventNode;
	
	switch ( debug ) {
	case 3:
	  System.out.println ( "ItemsetPrefixTree.add: added an event node " +
			       "with key: " + eventName + " - " +
			       eventKey + " in node: " + node );
	} // end switch debug
	
	// since we know there are no other nodes from here on
	// we could save time bye just creating nodes for the rest
	// of the items rather than checking above -kap
	
      } // end else we have to make a new node
      
    } // end for the event items

    // increment the size count... only if we are
    // not replacing an itemset of course
    if ( walker.itemset == null )
    {
      size++;
    } // end if there's no itemset there already

    // index the itemset
    walker.itemset = itemset;

    switch ( debug ) {
    case 3: case 2: //case 0:
      System.out.println ( "ItemsetPrefixTree.add: " +
			   "assigned itemset to node = " +
			   walker );
    } // end switch debug -kap
    
  } // END public void add ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Returns the itemset contained in the prefix tree
   * for a given itemset. Event items with the same name
   * are not considered unique so the returned itemset,
   * if any, may not contain the exact same items. This
   * is very useful for looking up conceptually duplicate
   * itemsets.
   *
   * @param itemset
   * the itemset to use to find the indexed itemset in
   * the prefix tree
   *
   * @return the itemset, null if not found
   */

  public ARMinerItemset getItemset ( ARMinerItemset itemset )
  {
    // get the node for the itemset
    HashNode node =  getNode ( itemset );

    // did we find it?
    if ( node == null )
    {
      return ( null );
    }
    
    return ( node.itemset );

  } // END public ARMinerItemset getItemset ( ARMinerItemset itemset )

//********************************************************//
 
  /**
   * Return the node for a given itemset.
   *
   * @param itemset  
   * the itemset for which we want to obtain the node
   *
   * @exception IllegalArgumentException  
   * <code>itemset</code> is null or is empty
   *
   * @return the node representing the itemset, null if not found
   */

  private HashNode getNode ( ARMinerItemset itemset )
  {
    if ( itemset == null || itemset.size() == 0 )
    {
      throw new IllegalArgumentException
	( "ItemsetPrefixTree.getNode(): " +
	  "aurgument must be non null and non empty" );
    } // end if itemset is stupid -kap

    Object obj;
    
    switch ( debug ) {
    case 3: case 2:
      System.out.println ( "ItemsetPrefixTree.getNode: Making new itemset" );
    case 1:
      System.out.println ( "ItemsetPrefixTree.getNode: " +
			   "of itemset: " + itemset );
    } // end switch debug -kap

    // it is not neccessary to set the number hash here since the 
    // constructor should get it from itemset -kap

    // start walking the tree at the root of course -kap
    HashNode walker = root;
    Integer key;

    // might need a string buffer to build event keys quickly
    StringBuffer sb = new StringBuffer();

    HashNode eventNode;

    // walk to the appropriate node for each regular item in the itemset -kap

    // start with the first item
    itemset.startItemIteration();

    while ( itemset.hasMoreItems() )
    {
      // the key is the item iteger representation -kap
      key = new Integer ( itemset.getNextItem() );
      
      // check if this item is an event
      if ( itemset.eventName ( key ) != null )
      {
	// then this is an event item and we should skip it for now
	continue;

      } // end if item is an event
      
      // if there is no node for this item then this itemset will not
      // be found in the tree -kap
      if ( ( obj = walker.children.get ( key ) ) == null )
      {
	return ( null );

      } // end if we're never gonna find the itemset

      // move to the node -kap
      walker = (HashNode)obj;

    } // end while more regular items -kap

    // need the event name to walk the tree.
    // remember to use event name as the key
    String eventName;

    // walk to the appropriate node for each event item in the itemset -kap

    // start with the first item again
    itemset.startItemIteration();

    while ( itemset.hasMoreItems() )
    {
      // the key is the item integer representation -kap
      key = new Integer ( itemset.getNextItem() );

      // get the event name
      eventName = itemset.eventName ( key );

      // check if this item is an event
      if ( eventName == null )
      {
	// then this is NOT an event item and we should skip it
	continue;

      } // end if item is NOT an event
      
      // if there is no node for this item then this itemset will not
      // be found in the tree
      // remember to use event name as the key
      if ( ( obj = walker.children.get ( eventName ) ) == null )
      {
	return ( null );

      } // end if we're never gonna find the itemset

      // Build the event key for it
      sb.setLength ( 0 );
      sb.append ( itemset.relativeBegin ( key ) ).
	append ( TSUtility.TIME_SEQUENCE_DELIMITER ).
	append ( itemset.relativeEnd ( key ) );
      
      eventNode = (HashNode)obj;

      switch ( debug ) {
      case 3: //case 0:
	System.out.println ( "ItemsetPrefixTree.getNode: walking to eventNode "
			     + eventName + " : " + eventNode );
      } // end switch debug

      // make sure it's there
      if ( ( obj = eventNode.children.get ( sb.toString() ) ) == null )
      {
	return ( null );
      } // end if we're never gonna find the itemset
      
      // move to the event node -kap
      walker = (HashNode)obj;
      
    } // end while more event items -kap

    switch ( debug ) {
    case 3: case 2: //case 0:
      System.out.println ( "ItemsetPrefixTree.getNode: Found node = " + 
			   walker );
    } // end switch debug -kap

    // now we just return the node
    return ( walker );

  } // END private HashNode getNode ( ARMinerItemset itemset )

//********************************************************//

  /**
   * Return the itemsets of the ItemsetPrefixTree. 
   *
   * @return
   * a vector containing the itemsets from the ItemsetPrefixTree
   */

  public Vector getItemsets() 
  { 
    Vector v = new Vector();

    // if the root has an itemset add it
    if ( root.itemset != null )
    {
      v.add ( root.itemset );

    } // end if root has itemset

    // get the rest of the itemsets
    getItemsets ( root, v );
    return ( v );

  } // END public Vector getItemsets() 

//********************************************************//

  /**
   * Return the itemsets of the ItemsetPrefixTree starting
   * from the specified node. 
   *
   * @param node
   * the node to start collecting itemsets from
   *
   * @param vector
   * the vector to add itemsets to
   */

  public void getItemsets ( HashNode node, Vector vector ) 
  { 
   if ( node.children.isEmpty() )
    {
      return;
      
    } // end if node children empty

    Enumeration e = node.children.keys();

    // key for regular items
    Object key;

    // temp node for checking for event node
    HashNode tempNode;

    while ( e.hasMoreElements() )
    {
      key = e.nextElement();
	
      // check if the node for key has an itemset
      tempNode = (HashNode)node.children.get ( key );
      
      // if there's an itemset in temp node add it
      if ( tempNode.itemset != null )
      {
	vector.add ( tempNode.itemset );
      } // end if temp node has an itemset

      // get the itemsets from this node
      getItemsets ( tempNode, vector );

    } // end while e has more
    
  } // END void Vector getItemsets ( HashNode node, Vector vector ) 

//********************************************************//
    
  /**
   * A private method which gets called recursively to
   * retrieve itemsets from each node of the
   * ItemsetPrefixTree and print them to a String.
   *
   * @param node  
   * node to traverse
   *
   * @param s  
   * StringBuffer in which we store the representation
   * of the nodes we saw so far
   */

  private void traversePrint ( HashNode node, StringBuffer sb )
  {
    level++;

    if ( node.children.isEmpty() )
    {
      level--;
      return;
      
    } // end if node children empty

    switch ( debug ) {
    case 3:
      System.out.println ( "ItemsetPrefixTree.traversePrint: " +
			   "children = " + node.children );
    } // end switch debug

    Enumeration e = node.children.keys();

    // key for regular items
    Object key;

    // key for event items
    String eventKey;

    // temp node for checking for event node
    HashNode tempNode, tempNode2;

    while ( e.hasMoreElements() )
    {
      key = e.nextElement();
      
      // make the nice indenting -kap
      for ( int i = 1; i < level; i++ )
      {
	sb.append ( "   " );
      }
      
      // check if the node for key is for an event item
      tempNode = (HashNode)node.children.get ( key );
      
      // changed to use all appends instead of + string operator -kap
      sb.append ( "<" );
      sb.append ( key );
      sb.append ( ">:[" );
      sb.append ( level );
      sb.append ( "] itemset= " );
      sb.append ( tempNode.itemset );
      sb.append ( "\n" );
      
      traversePrint ( tempNode, sb );
      
    } // end while e has more
    
    level--;
    
  } // END private void traversePrint ( HashNode node, StringBuffer sb )

//********************************************************//
	
  /**
   * Return a string representation of the ItemsetPrefixTree.
   *
   * @return string representation of ItemsetPrefixTree
   */

  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    traversePrint ( root, sb );
    return ( sb.toString() );

  } // END public String toString()

//********************************************************//
  
  /** 
   * Main method for testing purposes. -t option
   * specifies the cache file to load itemset from.
   */

  public static void main ( String[] args )
  {
    // if we have arguments we'll try to load a prefix tree
    // with all the itemsets in the cache file hopefully
    // specified by the arguments

    if ( args.length > 0 )
    {
      String cacheString =  "";
      
      try
      {
	// need the -t file too
	cacheString =  Utils.getOption ( 't', args );
      }
      catch ( Exception e ) 
      {
	e.printStackTrace();
	System.out.println ( "\n" + e.getMessage() );
	
      } //end catch
      
      // make the prefix tree
      ItemsetPrefixTree tree = new ItemsetPrefixTree();

      // start time
      long startTime = System.currentTimeMillis();
      
      // load each itemset
      try
      {
	// try making db cache reader
	DBCacheReader dbcache = new DBCacheReader ( cacheString );
	
	try
	{
	  while ( true )
	  {
	    tree.add ( dbcache.getNextItemset() );
	  }
	}
	catch ( EOFException e )
	{
	  dbcache.close();
	}
      }
      catch ( Exception e )
      {
	System.out.println ( e );
      }

      // end time
      long endTime = System.currentTimeMillis();
      long seconds = ( endTime / 1000 ) - ( startTime / 1000 );

      System.out.println ( "\nLoading Started : " + startTime );
      System.out.println ( "Loading Complete: " + endTime );
      System.out.println ( "time taken: " + seconds + " seconds\n" );
      System.out.println ( "to load " + tree.size() + " itemsets." );

      // all done
      return;

    } // end if args

    // - Manual testing - //

    // to test events and stuff
    Hashtable nh = new Hashtable();

    ItemsetPrefixTree tree = new ItemsetPrefixTree();

    ARMinerItemset is1 = new ARMinerItemset();
    is1.setNumberHash ( nh );
    ARMinerItemset is2 = new ARMinerItemset ( is1 );
    ARMinerItemset is3 = new ARMinerItemset ( is1 );
    ARMinerItemset is4 = new ARMinerItemset ( is1 );
    ARMinerItemset is5 = new ARMinerItemset ( is1 );
    ARMinerItemset is6 = new ARMinerItemset ( is1 );
    ARMinerItemset is7 = new ARMinerItemset ( is1 );
    ARMinerItemset is8 = new ARMinerItemset ( is1 );
    ARMinerItemset is9 = new ARMinerItemset ( is1 );

    nh.put ( new Integer ( 1 ), "odd-Event=1:10" );
    //nh.put ( new Integer ( 1 ), "odd-1" );
    is1.addItem ( 1 );
    //nh.put ( new Integer ( 2 ), "even-Event=2:20" );
    nh.put ( new Integer ( 2 ), "even-2" );
    is1.addItem ( 2 );
    nh.put ( new Integer ( 3 ), "odd-Event=3:30" );
    //nh.put ( new Integer ( 3 ), "odd-3" );
    is1.addItem ( 3 );
    //nh.put ( new Integer ( 4 ), "even-Event=4:40" );
    nh.put ( new Integer ( 4 ), "even-4" );
    is1.addItem ( 4 );
    is1.setSupport ( (float)0.4 );

    is2.addItem ( 1 );
    is2.addItem ( 2 );
    is2.addItem ( 3 );
    nh.put ( new Integer ( 7 ), "odd-Event=7:70" );
    //nh.put ( new Integer ( 7 ), "odd-7" );
    is2.addItem ( 7 );
    is2.setSupport ( (float)0.3 );

    is3.addItem ( 1 );
    is3.addItem ( 2 );
    is3.addItem ( 3 );
    nh.put ( new Integer ( 5 ), "odd-Event=5:50" );
    //nh.put ( new Integer ( 5 ), "odd-5" );
    is3.addItem ( 5 );
    //    nh.put ( new Integer ( 6 ), "even-Event=6:60" );
    nh.put ( new Integer ( 6 ), "even-6" );
    is3.addItem ( 6 );
    is3.setSupport ( (float)0.5 );

    is4.addItem ( 1 );
    is4.addItem ( 2 );
    is4.addItem ( 4 );
    is4.setSupport ( (float)0.65 );

    is5.addItem ( 1 );
    is5.addItem ( 2 );
    is5.addItem ( 5 );
    is5.setSupport ( (float)0.6 );

    is6.addItem ( 2 );
    is6.addItem ( 4 );
    is6.addItem ( 5 );
    is6.setSupport ( (float)0.55 );

    is7.addItem ( 1 );
    is7.addItem ( 2 );
    is7.setSupport ( (float)0.2 );

    is8.addItem ( 2 );
    is8.addItem ( 4 );
    is8.addItem ( 7 );
    //    nh.put ( new Integer ( 8 ), "even-Event=8:80" );
    nh.put ( new Integer ( 8 ), "even-8" );
    is8.addItem ( 8 );
    is8.setSupport ( (float)0.52 );

    is9.addItem ( 2 );
    is9.addItem ( 4 );
    is9.addItem ( 6 );
    is9.setSupport ( (float)0.59 );
    
    tree.add ( is2 );
    tree.add ( is3 );
    tree.add ( is1 );
    tree.add ( is4 );
    tree.add ( is5 );		
    tree.add ( is6 );
    tree.add ( is7 );
    tree.add ( is8 );
    tree.add ( is9 );

    System.out.println ( "The tree: \n" + tree + "\n" );

    System.out.println ( "is1 orig: " + is1 );
    System.out.println ( "is1 tree: " + tree.getItemset ( is1 ) );
    System.out.println ( "" );

    System.out.println ( "is2 orig: " + is2 );
    System.out.println ( "is2 tree: " + tree.getItemset ( is2 ) );
    System.out.println ( "" );

    System.out.println ( "is3 orig: " + is3 );
    System.out.println ( "is3 tree: " + tree.getItemset ( is3 ) );
    System.out.println ( "" );

    System.out.println ( "is4 orig: " + is4 );
    System.out.println ( "is4 tree: " + tree.getItemset ( is4 ) );
    System.out.println ( "" );

    System.out.println ( "is5 orig: " + is5 );
    System.out.println ( "is5 tree: " + tree.getItemset ( is5 ) );
    System.out.println ( "" );

    System.out.println ( "is6 orig: " + is6 );
    System.out.println ( "is6 tree: " + tree.getItemset ( is6 ) );
    System.out.println ( "" );

    System.out.println ( "is7 orig: " + is7 );
    System.out.println ( "is7 tree: " + tree.getItemset ( is7 ) );
    System.out.println ( "" );

    System.out.println ( "is8 orig: " + is8 );
    System.out.println ( "is8 tree: " + tree.getItemset ( is8 ) );
    System.out.println ( "" );

    System.out.println ( "is9 orig: " + is9 );
    System.out.println ( "is9 tree: " + tree.getItemset ( is9 ) );

    Vector v = tree.getItemsets();
    System.out.println ( "\nThe tree.getItemsets are: " + '\n' );

    for ( int i = 0; i < v.size(); i++ )
    {
      System.out.println ( v.get ( i ) );
    }

  } // END public static void main ( String[] args )

//********************************************************//

} // END public class ItemsetPrefixTree
