package wpi.timesequence;

import java.io.*;
import java.util.*;

import weka.core.*;
import weka.filters.*;

/** 
 * Finds time periods during which events occur in time
 * sequence attribute values. Three events are currently
 * supported: increase, decrease and sustain.
 * A new attribute containing the events found is appended
 * after the last attribute.
 *
 * @author Keith A. Pray (kap@wpi.edu)
 * @version $Revision: 0.9 $
 */

public class AverageSequenceFilter extends EventFilter
{

//********************************************************//
// ----- Data Members -----
//********************************************************//  

  /** The string to append to an attribute's name for
      average */
  public static String AVERAGE = "-average";

//********************************************************//
// ----- Methods-----
//********************************************************//

  /**
   * Computes the average value of the time sequence
   * attribute specified for the given instance.
   */

  public double GetAverage ( Instance instance,
			     String attribute )
  {
    // get the values
    double [] values = TSUtility.GetValues ( instance, attribute );

    double total = 0;

    for ( int i = 0; i < values.length; i++ )
    {
      total += values[i];

    } // end for each value

    return ( total / values.length );

  } // END public double GetAverage ( Instance instance,
  //			     String attribute )

//********************************************************//

  /**
   * Sets the output format. Takes the buffered input
   * instances and generates valid strings for the new
   * attributes.
   */
  
  private void setOutputFormat() 
  {
    // make a new instances for the output format
    Instances outputFormat = new Instances ( getInputFormat(), 0 );

    // if attributes were specified just do those
    // otherwise we should try to find events in all the time
    // sequence attributes
    if ( timeSequenceAttributeVector == null )
    {
      FindTimeSequenceAttributes();

    } // end if attributes not set

    for ( int i = 0; i < timeSequenceAttributeVector.size(); i++ )
    {
      outputFormat.
	insertAttributeAt ( new Attribute ( timeSequenceAttributeVector.
					    elementAt ( i ) + 
					    AVERAGE ),
			    outputFormat.numAttributes() );
      
    } // end for each time sequence attribute

    setOutputFormat ( outputFormat );
    
  } // END private void setOutputFormat() 

 //********************************************************//
  
  /**
   * Signifies that this batch of input to the filter is
   * finished. 
   *
   * @return true
   * if there are instances pending output.
   *
   * @exception IllegalStateException
   * if no input structure has been defined.
   */

  public boolean batchFinished() 
  {
    // make sure we have an input format
    if ( getInputFormat() == null ) 
      {
	throw new IllegalStateException ( "No input instance format defined" );
      }
    
    // if the output format is not defined then set it and
    // generate the output instance
    if ( ! isOutputFormatDefined() ) 
      {
	setOutputFormat();

	// go through all the buffered instances
	for ( int i = 0; i < getInputFormat().numInstances(); i++ )
	{	  	
	  // make a new instance with the output data set
	  Instance newInstance = new Instance ( getOutputFormat().
						numAttributes() );
	  newInstance.setDataset ( getOutputFormat() );
	  
	  // set all the attributes of this new instance
	  // with values from a buffered input instance
	  for ( int a = 0; a < getInputFormat().numAttributes(); a++ )
	  {
	    if ( getOutputFormat().attribute ( a ).isNumeric() )
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     value ( a ) );

	    }  // end if numeric set value using value
	    else
	    {
	      newInstance.setValue ( a, getInputFormat().instance ( i ).
				     stringValue ( a ) );
	    } // end else set using stringValue

	  } // end for all input attributes

	  // get the events found
	  for ( int e = 0; e < timeSequenceAttributeVector.size(); e++ )
	  {
	    newInstance.
	      setValue ( getOutputFormat().
			 attribute ( timeSequenceAttributeVector.
				     elementAt ( e ) + 
				     AVERAGE ),
			 GetAverage ( getInputFormat().instance ( i ),
				      (String)timeSequenceAttributeVector.
				      elementAt ( e ) ) );
	    
	  } // end for each attribute to find events in

	  // all set, push her on out
	  push ( newInstance );
	    
	} // end for all input instances
	
      } // end if output format not defined
    
    flushInput();
    m_NewBatch = true;
    return ( numPendingOutput() != 0 );

  } // END public boolean batchFinished() 

//********************************************************//

  /**
   * Main method for testing this class.
   *
   * @param argv
   * should contain arguments to the filter: 
   * use -h for help
   */

  public static void main ( String [] argv ) 
  {
    try 
      {
	if ( Utils.getFlag ( 'b', argv ) ) 
	  {
	    Filter.batchFilterFile ( new AverageSequenceFilter(), argv ); 
	  } 
	else 
	  {
	    Filter.filterFile ( new AverageSequenceFilter(), argv );
	  }
      } 
    catch ( Exception ex ) 
      {
	System.out.println ( ex.getMessage() );
      }

  } // END public static void main ( String [] argv )

//********************************************************//
 
} // END public class AverageSequenceFilter extends EventFilter
